"""Class which tests ColumnSelectionPandasChecker."""
import pylint.testutils
import astroid
import dslinter.plugin


class TestColumnSelectionPandasChecker(pylint.testutils.CheckerTestCase):
    """Class which tests ColumnSelectionPandasChecker."""

    CHECKER_CLASS = dslinter.plugin.ColumnSelectionPandasChecker

    def test_column_not_selected(self):
        """A message should be added if there is no column selection after the dataframe is imported."""
        script = """
        import pandas as pd
        df = pd.read_csv('data.csv')
        """
        module = astroid.parse(script)
        assign_node = module.body[-1]
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="column-selection-pandas", node=assign_node)):
            self.checker.visit_module(module)

    def test_column_selected1(self):
        """No message should be added if there is column selection after the dataframe is imported."""
        script = """
        import pandas as pd
        df = pd.read_csv('data.csv')
        df = df[['col1', 'col2', 'col3']]
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)

    def test_column_selected2(self):
        """No message should be added if there is column selection after the dataframe is imported."""
        script = """
        import pandas as pd
        df = pd.read_csv('data.csv')
        print(df)
        df = df[['col1', 'col2', 'col3']]
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)

    def test_a_column_not_selected(self):
        """A message should be added if some imported dataframe are not selected by columns."""
        script = """
        import pandas as pd
        df1 = pd.read_csv('data1.csv')
        df2 = pd.read_csv('data2.csv')
        df1 = df1[['col1', 'col2', 'col3']]
        """
        module = astroid.parse(script)
        assign_node = module.body[-2]
        with self.assertAddsMessages(pylint.testutils.MessageTest(msg_id="column-selection-pandas", node=assign_node)):
            self.checker.visit_module(module)

    def test_all_columns_selected(self):
        """No message should be added if every imported dataframe is selected by columns."""
        script = """
        import pandas as pd
        df1 = pd.read_csv('data1.csv')
        df2 = pd.read_csv('data2.csv')
        df1 = df1[['col1', 'col2', 'col3']]
        df3 = df2[['col1', 'col2']]
        """
        module = astroid.parse(script)
        with self.assertNoMessages():
            self.checker.visit_module(module)



"""Checker which checks column is selected after the dataframe is imported."""
from collections import defaultdict

import astroid
from pylint.checkers import BaseChecker
from pylint.interfaces import IAstroidChecker

from dslinter.utils.exception_handler import ExceptionHandler


class ColumnSelectionPandasChecker(BaseChecker):
    """Checker which checks column is selected after the dataframe is imported."""

    __implements__ = IAstroidChecker

    name = "column-selection-pandas"
    priority = -1
    msgs = {
        "R5504": (
            "There is no column selection after the dataframe is imported.",
            "column-selection-pandas",
            "Column should be selected after the dataframe is imported for better readability."
        )
    }
    options = ()

    _has_column_selection = False
    _dataframe_selected_dict = defaultdict(bool)
    _name_node_dict = defaultdict()
    _data_import_functions = ["read_csv", "read_table", "read_fwf", "read_excel", "read_xml"]

    def visit_module(self, module: astroid.Module):
        """
        Visit module and see whether the rule is violated.
        :param module:
        :return:
        """
        try:
            for idx, node in enumerate(module.body):
                if isinstance(node, astroid.Assign):
                    # check if there is a imported dataframe
                    if(
                        hasattr(node.value, "func")
                        and hasattr(node.value.func, "attrname")
                        and node.value.func.attrname in self._data_import_functions
                        and hasattr(node.value.func, "expr")
                        and hasattr(node.value.func.expr, "name")
                        and node.value.func.expr.name in ["pandas", "pd"]
                        and len(node.targets) > 0
                        and hasattr(node.targets[0], "name")
                    ):
                        _imported_dataframe_name = node.targets[0].name
                        self._dataframe_selected_dict[_imported_dataframe_name] = False
                        self._name_node_dict[_imported_dataframe_name] = node
                    # check if dataframe columns are selected
                    if(
                        hasattr(node, "value")
                        and isinstance(node.value, astroid.Subscript)
                        and hasattr(node.value, "value")
                        and hasattr(node.value.value, "name")
                        and node.value.value.name in self._dataframe_selected_dict
                        and hasattr(node.value, "slice")
                        and isinstance(node.value.slice, astroid.nodes.List)
                    ):
                        _selected_dataframe_name = node.value.value.name
                        self._dataframe_selected_dict[_selected_dataframe_name] = True

            # check which dataframe is imported but not selected by columns
            for k, v in self._dataframe_selected_dict.items():
                if v is False:
                    self.add_message("column-selection-pandas", node=self._name_node_dict[k])
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, module)



"""Checker which checks whether random seed is set in tensorflow"""
from pylint.interfaces import IAstroidChecker
from pylint.checkers import BaseChecker
import astroid

from dslinter.utils.exception_handler import ExceptionHandler
from dslinter.utils.randomness_control_helper import check_main_module, has_import


class RandomnessControlTensorflowChecker(BaseChecker):
    """Checker which checks whether random seed is set in tensorflow"""
    __implements__ = IAstroidChecker

    name = "randomness-control-tensorflow"
    priority = -1
    msgs = {
        "W5510": (
            "tf.random.set_seed() is not set in TensorFlow program",
            "randomness-control-tensorflow",
            "tf.random.set_seed() should be set in TensorFlow program for reproducible result"
        )
    }
    options = (
        (
            "no_main_module_check_randomness_control_tensorflow",
            {
                "default": False,
                "type": "yn",
                "metavar": "<y_or_n>",
                "help": "Check every module whether tf.random.set_seed() is used.",
            },
        ),
    )

    _import_tensorflow = False
    _has_manual_seed = False

    def visit_import(self, node: astroid.Import):
        """
        Check whether there is a tensorflow import.
        :param node: import node
        """
        try:
            if self._import_tensorflow is False:
                self._import_tensorflow = has_import(node, "tensorflow")
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, node)

    def visit_module(self, module: astroid.Module):
        """
        Check whether there is a rule violation.
        :param module:
        """
        try:
            _is_main_module = check_main_module(module)
            if self.config.no_main_module_check_randomness_control_tensorflow is False and _is_main_module is False:
                return

            for node in module.body:
                if isinstance(node, astroid.nodes.Expr) and hasattr(node, "value"):
                    call_node = node.value
                    if(
                        hasattr(call_node, "func")
                        and hasattr(call_node.func, "attrname")
                        and call_node.func.attrname == "set_seed"
                        and hasattr(call_node.func.expr, "attrname")
                        and call_node.func.expr.attrname == "random"
                        and hasattr(call_node.func.expr, "expr")
                        and hasattr(call_node.func.expr.expr, "name")
                        and call_node.func.expr.expr.name in ["tf", "tensorflow"]
                    ):
                        self._has_manual_seed = True

            if(
                self._import_tensorflow is True
                and self._has_manual_seed is False
            ):
                self.add_message("randomness-control-tensorflow", node=module)
        except: # pylint: disable = bare-except
            ExceptionHandler.handle(self, module)
